#! /usr/bin/python

if __name__ == '__main__':
    import pygtk
    pygtk.require("2.0")

import time, cPickle
import gobject, gtk, gtk.glade
from watcher import Watcher

def time_string(latency):
    if latency == None:
        return "?"
    latency = int(latency)
    hours = latency / 3600
    latency -= hours * 3600
    minutes = latency / 60
    latency -= minutes * 60
    seconds = latency
    latency = ''
    if hours:
        latency = '%dh' % hours
    if hours or minutes:
        latency = latency + '%dm' % minutes
    latency = latency + '%ds' % seconds
    return latency

class WatcherGUI:
    def __init__(self, watcher):
        self.watcher = watcher
        watcher.gui = self

        self.done = 0

        xml = gtk.glade.XML('gtkwatcher.glade')
        self.xml = xml

        # main menu
        xml.signal_connect('do_poll', self.do_poll)
        xml.signal_connect('do_exit', self.do_quit)

        # source panel
        self.src_popup = xml.get_widget("src_popup")
        xml.signal_connect('do_src_abandon', self.do_src_abandon)
        xml.get_widget("source_message_options1").set_sensitive(0)
        self.src_age_item = xml.get_widget("src_age_item")
        self.src_age_item.set_sensitive(0)
        self.src_abandon_item = xml.get_widget("src_abandon_item")

        # dest panel
        self.dst_popup = xml.get_widget("dst_popup")
        xml.get_widget("dest_message_options1").set_sensitive(0)
        self.dst_sent_item = xml.get_widget("dst_sent_item")
        self.dst_sent_item.set_sensitive(0)
        self.dst_original_item = xml.get_widget("dst_original_item")
        self.dst_flush_item = xml.get_widget("dst_flush_item")

        xml.signal_connect('do_dst_flush', self.do_dst_flush)
        xml.signal_connect('do_dst_original', self.do_dst_original)

        # panel contents
        self.src_model = gtk.ListStore(gobject.TYPE_STRING,
                                       gobject.TYPE_STRING,
                                       gobject.TYPE_PYOBJECT)
        self.dst_model = gtk.ListStore(gobject.TYPE_STRING,
                                       gobject.TYPE_STRING,
                                       gobject.TYPE_PYOBJECT)

        view = xml.get_widget('src_treeview')
        view.connect("button_press_event", self.do_src_popup)
        view.set_model(self.src_model)
        r = gtk.CellRendererText()
        view.append_column(gtk.TreeViewColumn("Message ID", r, text=0))
        view.append_column(gtk.TreeViewColumn("Message Sent", r, text=1))
        sel = view.get_selection()
        sel.set_mode(gtk.SELECTION_SINGLE)
        sel.connect("changed", self.do_src_select)
        self.src_sel = sel

        view = xml.get_widget('dst_treeview')
        view.connect("button_press_event", self.do_dst_popup)
        view.set_model(self.dst_model)
        r = gtk.CellRendererText()
        view.append_column(gtk.TreeViewColumn("Message ID", r, text=0))
        view.append_column(gtk.TreeViewColumn("Latency", r, text=1))
        sel = view.get_selection()
        sel.set_mode(gtk.SELECTION_SINGLE)
        sel.connect("changed", self.do_dst_select)
        self.dst_sel = sel

        self.srcwin = xml.get_widget('srcwin')
        self.dstwin = xml.get_widget('dstwin')
        #self.src_clist.connect('select-row', self.do_src_select)
        #self.dst_clist.connect('select-row', self.do_dst_select)
        self.text = xml.get_widget('text1')
        self.text.set_wrap_mode(gtk.WRAP_NONE)
        self.textwin = xml.get_widget('textwin')
        xml.get_widget('window1').set_size_request(500,300)
        self.do_update()

    def do_quit(self, widget):
        print "doing quit"
        self.done = 1
        # mainquit asserts, because we aren't actually in a mainloop
        #gtk.mainquit()

    def update_text(self, text, skipHeaders=0):
        buf = self.text.get_buffer()
        buf.set_text(text)
        # now make the end of the buffer visible
        # XXX: this flashes. They removed freeze/thaw.. how to fix?
        # XXX: if skipHeaders, find the first blank line and put that at top
        iter = buf.get_iter_at_line(-1)
        #print iter.get_line()
        # turn it into a mark, as scroll_to_iter depends upon height
        # calculations that are done in an idle task, so it won't get it right
        # until later
        mark = buf.create_mark("end", iter, 0)
        if skipHeaders:
            self.text.scroll_to_mark(mark, within_margin=0)

    def do_src_select(self, sel):
        model, iter = sel.get_selected()
        if not iter:
            return  # deselected
        m = model.get_value(iter, 2)
        # get the message text from the Outstanding list
        text = m.data
        self.update_text(text)
        # need to deselect the one in the other list so we can sense when it
        # becomes reselected
        self.dst_sel.unselect_all()

    def do_src_popup(self, view, event):
        if event.button != 3:
            return
        pathset = view.get_path_at_pos(event.x, event.y)
        if pathset:
            path, viewcol, cell_x, cell_y = pathset
            iter = self.src_model.get_iter(path)
            m = self.src_model.get_value(iter, 2)
            age = self.watcher.age(m.msgid)
            label = self.src_age_item.get_child()
            label.set_text("Age[%d]: %s" % (m.msgid, time_string(age)))
            label = self.src_abandon_item.get_child()
            label.set_text("Abandon Message [%d]" % m.msgid)
            self.src_abandon_item.set_sensitive(1)
        else:
            label = self.src_age_item.get_child()
            label.set_text("Age: --")
            label = self.src_abandon_item.get_child()
            label.set_text("Abandon Message")
            self.src_abandon_item.set_sensitive(0)
        self.src_popup.popup(None, None, None, event.button, event.time)

    def do_src_abandon(self, menuitem):
        # which message? find the selection
        model, iter = self.src_sel.get_selected()
        if not iter:
            print "abandon, no iter!"
            return
        m = model.get_value(iter, 2)
        print "abandon msgid", m.msgid
        self.watcher.abandon(m.msgid)
        self.do_update()

    def do_dst_popup(self, view, event):
        if event.button != 3:
            return
        pathset = view.get_path_at_pos(event.x, event.y)
        if pathset:
            path, viewcol, cell_x, cell_y = pathset
            iter = self.dst_model.get_iter(path)
            m = self.dst_model.get_value(iter, 2)
            txtime = self.watcher.txtime(m.msgid)
            sent = time.strftime("%H:%M   %d %b %Y", time.localtime(txtime))
            label = self.dst_sent_item.get_child()
            label.set_text("Sent[%d]: %s" % (m.msgid, sent))
            self.dst_original_item.set_sensitive(1)
            label = self.dst_original_item.get_child()
            label.set_text("See Original [%d]" % m.msgid)
            self.dst_flush_item.set_sensitive(1)
            label = self.dst_flush_item.get_child()
            label.set_text("Flush Message [%d]" % m.msgid)
        else:
            label = self.dst_sent_item.get_child()
            label.set_text("Sent: --")
            self.dst_original_item.set_sensitive(0)
            label = self.dst_original_item.get_child()
            label.set_text("See Original")
            self.dst_flush_item.set_sensitive(0)
            label = self.dst_flush_item.get_child()
            label.set_text("Flush Message")
        self.dst_popup.popup(None, None, None, event.button, event.time)

    def do_dst_flush(self, menuitem):
        # which message? find the selection
        model, iter = self.dst_sel.get_selected()
        if not iter:
            return
        m = model.get_value(iter, 2)
        print "flush msgid", m.msgid
        self.watcher.flush(m.msgid)
        self.do_update()

    def do_dst_original(self, menuitem):
        # which message? find the selection
        model, iter = self.dst_sel.get_selected()
        if not iter:
            return
        dst_msg = model.get_value(iter, 2)
        src_msg = self.watcher.source.msgs.get(dst_msg.msgid, None)
        if src_msg:
            text = src_msg.data
            self.update_text(text, skipHeaders=0)
        #self.dst_sel.unselect_all()

    def do_dst_select(self, sel):
        model, iter = sel.get_selected()
        if not iter:
            return
        m = model.get_value(iter, 2)
        text = m.data
        self.update_text(text, skipHeaders=1)
        self.src_sel.unselect_all()

    def do_poll(self, *args):
        self.watcher.poll()
        return 1
    def do_update(self):
        src = self.src_model
        src.clear()
        for msg, txtime in self.watcher.outstanding():
            #elapsed = time_string(int(time.time()) - txtime)
            sent = time.strftime("%H:%M   %d %b %Y", time.localtime(txtime))
            iter = src.append()
            src.set(iter,
                    0, msg.msgid,
                    1, sent,
                    2, msg)
            #ol.set_row_data(row, msgid)
            #ol.set_selectable(row, 0)
        #ol.moveto(maxrow, 0, 1.0, 0.0)

        dst = self.dst_model
        dst.clear()
        for msg, latency in self.watcher.received():
            iter = dst.append()
            dst.set(iter,
                    0, msg.msgid,
                    1, time_string(latency),
                    2, msg)
            #rl.set_row_data(row, msgid)
            #rl.set_selectable(row, 0)
        #rl.moveto(maxrow, 0, 1.0, 0.0)

class GtkWatcher(Watcher):
    def checker(self, m):
        self.gui.do_update()
    def start(self):
        self.gui = WatcherGUI(self)
        Watcher.start(self)
    def stop(self):
        self.gui = None
        Watcher.stop(self)

def makeWatcher():
    from gtkwatcher import GtkWatcher
    from watcher import DirWatcher, NewsWatcher
    import sys, getopt
    w = GtkWatcher()
    usage = """
    gtkwatcher.py
      --source sdir
      [--dest ddir]...
      --nntp host:port:user:pass [--group a]..
    """
    try:
        opts, args = getopt.getopt(sys.argv[1:], '',
                                   ['source=', 'dest=',
                                    'nntp=', 'group=', 'tag='])
    except getopt.GetoptError:
        print usage
        sys.exit(-1)
    nntp = None
    groups = []
    tag = None
    for key, value in opts:
        if key == "--source":
            print "adding source maildir '%s'" % value
            w.addSource(DirWatcher(value))
        if key == "--dest":
            print "adding dest maildir '%s'" % value
            w.addDest(DirWatcher(value))
        if key == "--nntp":
            nntp = value
        if key == "--group":
            groups.append(value)
        if key == "--tag":
            tag = value
    if groups and not nntp:
        raise "Must have --nntp server specified to use --group"
    if tag and not nntp:
        raise "Must have --nntp set to use message filtering tag"
    if nntp and not groups:
        raise "Gave NNTP server but no groups to poll: what's the point?"
    if nntp:
        # parse server spec
        fields = nntp.split(':')
        fields.extend([None] * (4 - len(fields)))
        host, port, user, passwd = fields
        if port:
            port = int(port)
        nw = NewsWatcher(host, groups,
                         user, passwd, port, tag)
        w.addDest(nw)
    return w

def main():
    try:
        f = open("watcher.pickle", "r")
        w = cPickle.load(f)
        f.close()
    except:
        print "unable to load pickle"
        w = makeWatcher()
    print "running"
    w.start()
    while not w.gui.done:
        gtk.mainiteration()
    print "done"
    w.stop()
    # save pickle of seen messages
    f = open("watcher.pickle", "w")
    cPickle.dump(w, f, 1)
    f.close()

if __name__ == '__main__': main()



#! /usr/bin/python

import GnuPGInterface
import string, re, sys

gnupg = GnuPGInterface.GnuPG()
gnupg.options.extra_args += ['--homedir', 'remailer-keys', '--batch']

# The keyring should hold private keys for all the remailers in the chain,
# and each one should have a passphrase equal to the name of the key.

def passphrase(bigname):
    return re.search(r'(\w+@\w+\.\w+)', bigname).group(1)

def unwind(recipient, message):
    """Unwind a message. Takes two parameters: the To: address of the
    message and the message body itself.

    This function will return a list of steps. Each element is a dict
    describing what happened on that particular step. The decryption and
    remailing are separate steps. The dict keys are as follows:

     recipient: the target of that step, as requested by the previous step
     encrypted: 1 if the 'Encrypted: PGP' flag was set, else non-existent
     anon-to: name in the Anon-To: header, if present
     subject: subject specified in a ## header, if present
     message: plaintext of the message (for the last step)

    Typical steps:
     recipient=rem1, encrypted=1  ->recurse BRANCH2
     recipient=rem1, anon-to=rem2  BRANCH3, recurse
     recipient=rem2, encrypted=1  ->recurse BRANCH2
     recipient=rem2, anon-to=user, subject=test  BRANCH3, recurse
     recipient=user, message=plaintext  : BRANCH1, terminate

    It throws an exception if anything about the message is incorrect:
     message doesn't start with ::\nEncrypted: PGP
     message isn't encrypted or is encrypted to the wrong key (not TO)
     decrypted message doesn't start with ::\nAnon-To:
    """
    # the encrypted message is handled by unwind1. It will look like:
    # ::\nEncrypted: PGP\n\n-----BEGIN PGP MESSAGE----\n etc

    # once decrypted, the message should look like:
    # ::\nAnon-To: dest@foo.com\n\n##\nSubject: test subject\n\nmessage body

    if 0:
        print "-----"
        print message
        print "-----"

    step = {}

    # see if we should end the recursion now
    if not re.search(r'^rem\d@test.test', recipient):
        # BRANCH 1
        # plaintext message
        step['recipient'] = recipient
        step['message'] = message
        return [step]
    message = string.split(message, "\n")
    if message[0] != '::':
        raise 'bad message', 'first line was not ::'
    if message[2] != '':
        raise 'bad message', 'third line was not blank'
    # see if it's encrypted
    if message[1] == 'Encrypted: PGP':
        # yes, decrypt it and recurse
        # BRANCH2
        # step is: recipient=rem1, encrypted=1
        crypttext = message[3:]
        assert(message[3] == '-----BEGIN PGP MESSAGE-----')
        # decrypt here
        gnupg.passphrase = passphrase(recipient) # passphrase == keyname
        p = gnupg.run(['--decrypt'],
                      create_fhs=['stdin', 'stdout'],
                      )
        p.handles['stdin'].write(string.join(crypttext,"\n"))
        p.handles['stdin'].close()
        plaintext = p.handles['stdout'].read()
        p.handles['stdout'].close()
        p.wait()
        step['recipient'] = recipient
        step['encrypted'] = 1
        return [step] + unwind(recipient, plaintext)

    # BRANCH3

    # message is now:
    #  ::\nAnon-To: remN@test.set (or dest@foo.com)\n\nmessage body
    # or
    #  ::\nAnon-To: dest@foo.com\n\n##\nSubject: subject\n\nmessage body

    r = re.search(r'^Anon-To: (.*)$', message[1])
    if not r:
        print "Bad Message, no Anon-To"
        print message
        raise 'bad message', "no anon-to"
    step['recipient'] = recipient
    step['anon-to'] = r.group(1)

    message = message[3:]
    # now: "message body", or "##\nSubject: sub\n\nmessage body"

    if message[0] == '##':
        # there are ## headers included
        r = re.search(r'^Subject: (.*)$', message[1])
        if not r:
            raise 'bad message', "## but no Subject: header"
        step['subject'] = r.group(1)
        if message[2] != '':
            raise 'bad message', "no blank line after ## headers"
        message = message[3:]
    # now: "message body"
    message = string.join(message, "\n")
    return [step] + unwind(step['anon-to'], message)

def insistEquals(one, two):
    if one != two:
        raise "results don't match expected", "'%s' != '%s'" % (one, two)


def test_chain(firsthop, crypttext, plaintext, recipient, chain, subject=None):
    """Verify that the crypttext message does indeed match the plaintext
    message, sent to a given recipient and encrypted to the given remailer
    chain."""

    # chain is a list of remailers with long names: rem1@test.test, etc

    # build up the list of what we expect to see at each step
    expected_chain = []
    for i in range(len(chain)-1):
        expected_chain.append({'recipient': chain[i],
                               'encrypted': 1,
                               })
        expected_chain.append({'recipient': chain[i],
                               'anon-to': chain[i+1],
                               })
    last = chain[len(chain)-1]
    expected_chain.append({'recipient': last,
                           'encrypted': 1,
                           })
    pentultimate = {'recipient': last,
                    'anon-to': recipient,
                    }
    if subject:
        pentultimate['subject'] = subject
    expected_chain.append(pentultimate)
    expected_chain.append({'recipient': recipient,
                           'message': plaintext,
                           })

    # unwind the messge
    insistEquals(firsthop, chain[0])
    results = unwind(firsthop, crypttext)

    # compare against expectations
    insistEquals(len(expected_chain), len(results))
    for i in range(len(results)):
        #print i, results[i], expected_chain[i]
        insistEquals(results[i], expected_chain[i])
    print "TEST CASE PASSED"



def test1():
    m1 = open("m1").read()
    chain = unwind("rem3@test.test", m1)
    for link in chain: print link

def test2():
    m1 = open("m1").read()
    chain = ["rem3", "rem2", "rem1", "rem1"]
    test_chain(firsthop="rem3@test.test", crypttext=m1,
               plaintext="test message\n",
               recipient="warner@lothar.com",
               chain=chain,
               subject="test subject")

def main():
    # argv is: ['recipient', 'chain1,chain2,chain3', 'subject']
    # plaintext is always "test message\n"
    # crypttext arrives on stdin
    recipient = sys.argv[1]
    chain = string.split(sys.argv[2], ',')
    subject = sys.argv[3]
    crypttext = sys.stdin.read()
    test_chain(firsthop=chain[0],
               crypttext=crypttext,
               plaintext="test message\n",
               recipient=recipient,
               chain=chain,
               subject=subject)


if __name__ == '__main__':
    main()





#! /usr/bin/python

import popen2, commands, re, os, sys

# create the keyrings. There are two: the full rings are used to create the
# test cases, then the stripped-down set are used when running the tests.

# The full ring contains the following keys (both public and secret):
#  (with the given signatures)

# #1: owner1 <user@test>
# #2: owner2 <user@test>
#      signed by owner1 ("careful checking 3")
# #3: other <other@test>
#      signed by owner1 ("casual checking 2")
# #4: unknown <unknown@test>
# #5: trusted <trusted@test>
#      signed by other ("2")
# #6: untrusted <untrusted@test>
#      signed by trusted ("2")

# The stripped ring contains the following subset, with the given trust values:

# owner1 (public+secret) ultimate  u
# owner2 (public+secret) ultimate  u
# other (public only) marginal     m
# trusted (public only) do not trust n
# untrusted (public only) "don't know" q

# as a result, messages signed by these keys should have the following computed
# trust values:
#  owner1 TRUST_ULTIMATE u
#  owner2 TRUST_ULTIMATE u
#  other  TRUST_FULLY f
#  unknown none -
#  trusted TRUST_MARGINAL m
#  untrusted TRUST_NONE -

# messages signed by owner1 or owner2 should be accepted with TRUST_ULTIMATE,
#  because we have secret keys for them
# messages signed by other should have

homedir = sys.argv[1]

def make_key(name):
    # this requires a lot of random bytes. look for a gpg option to suck
    # on /dev/urandom instead of the blocking /dev/random
    cmd = "gpg --homedir %s --batch --gen-key 2>&1" % homedir
    (stdout,stdin) = popen2.popen2(cmd)
    stdin.write("Key-Type: DSA\n")
    stdin.write("Key-Length: 1024\n")
    stdin.write("Subkey-Type: ELG-E\n")
    stdin.write("Subkey-Length: 768\n")
    stdin.write("Name-Real: %s\n" % name)
    stdin.write("Name-Email: %s@test\n" % name)
    stdin.write("Passphrase: %s\n" % name)
    stdin.write("%commit\n")
    stdin.write("%echo done\n")
    stdin.close()
    while 1:
        out = stdout.readline()
        print out,
        if out == "gpg: done\n": break
def get_keyid(name):
    # now what keyid did it get?
    cmd = "gpg --homedir %s --with-colons --fingerprint %s" % (homedir,name)
    (s,out) = commands.getstatusoutput(cmd)
    assert(s==0)
    r = re.search(r'^fpr:::::::::(\w+):',out, re.M)
    if r == None:
        print "problem, out '%s'" % out
        return None
    id = r.group(1) # long form
    return id

# gpg claims that signing keys in batch mode is not possible. To work around
# it would probably involve a lot of expect-style futzing.
def sign_key(signer, key, checklevel):
    cmd = "gpg --homedir %s --local-user %s --sign-key %s" % \
          (homedir, signer, key)
    print cmd
    print " check level %d" % checklevel

# this can probably be automated by emitting a new copy of trustdb.gpg
def make_trustdb(trust):
    print "updating trustdb"
    cmd = "gpg --homedir %s --import-ownertrust" % homedir
    (stdout,stdin) = popen2.popen2(cmd)
    for id in trust.keys():
        stdin.write("%s:%d:\n" % (id, trust[id]))
    stdin.close()


users = ("owner1", "owner2", "other", "unknown", "trusted", "untrusted")
if 1:
    os.mkdir(homedir)
    os.chmod(homedir, 0700)
    for u in users:
        print "creating key for", u
        make_key(u)
os.chmod(os.path.join(homedir,"pubring.gpg"), 0600)
os.chmod(os.path.join(homedir,"secring.gpg"), 0600)
os.chmod(os.path.join(homedir,"trustdb.gpg"), 0600)

# get keyids
id = {}
for u in users:
    id[u] = get_keyid(u)

# now sign them
sign_key("owner1", "owner2", 3)
sign_key("owner1", "other", 2)
sign_key("other", "trusted", 2)
sign_key("trusted", "untrusted", 2)

# and assign trust values
trust = {
    id["owner1"]: 6,    # u
    id["owner2"]: 6,    # u
    id["other"]: 4,     # m
    id["trusted"]: 3,    # n
    id["untrusted"]: 2, # q
    }

make_trustdb(trust)

# update_trustdb



#! /usr/bin/python

import os, commands, re, time
import GnuPGInterface

# emit test case files into test-cases/*

# keys:
#  1: owner1 <user@test>
#  2: owner2 <user@test>
#  3: other <other@test>
#  4: unknown key

homedir = "gpg-keys/full-rings"
testcasedir = "gpg-testcases"
users = ("owner1", "owner2", "other", "unknown", "trusted", "untrusted")

def get_keyid(name):
    # now what keyid did it get?
    cmd = "gpg --homedir %s --with-colons --fingerprint %s" % (homedir,name)
    (s,out) = commands.getstatusoutput(cmd)
    assert(s==0)
    r = re.search(r'^fpr:::::::::(\w+):',out, re.M)
    if r == None:
        print "problem, out '%s'" % out
        return None
    id = r.group(1) # long form
    return id

class Encrypt(GnuPGInterface.GnuPG):
    def __init__(self):
        GnuPGInterface.GnuPG.__init__(self)
        self.setup_my_options()

    def setup_my_options(self):
        self.options.armor = 1
        self.options.meta_interactive = 0
        self.options.extra_args.append('--no-secmem-warning')
        self.options.homedir = homedir

    def encrypt_string(self, string, recipients):
        self.options.recipients = recipients   # a list!
        proc = self.run(['--encrypt'], create_fhs=['stdin', 'stdout'])
        proc.handles['stdin'].write(string)
        proc.handles['stdin'].close()
        output = proc.handles['stdout'].read()
        proc.handles['stdout'].close()
        proc.wait()
        return output

    def signencrypt_string(self, string, signer, passphrase, recipients):
        self.options.recipients = recipients   # a list!
        self.options.default_key = signer
        self.passphrase = passphrase
        proc = self.run(['--sign', '--encrypt'],
                        create_fhs=['stdin', 'stdout'])
        proc.handles['stdin'].write(string)
        proc.handles['stdin'].close()
        output = proc.handles['stdout'].read()
        proc.handles['stdout'].close()
        proc.wait()
        return output

    def sign_string(self, string, signer, passphrase):
        self.options.recipients = []
        self.options.default_key = signer
        self.passphrase = passphrase
        proc = self.run(['--sign'],
                        create_fhs=['stdin', 'stdout'])
        proc.handles['stdin'].write(string)
        proc.handles['stdin'].close()
        output = proc.handles['stdout'].read()
        proc.handles['stdout'].close()
        proc.wait()
        return output

    def sym_string(self, string, passphrase):
        self.options.recipients = []
        self.passphrase = passphrase
        proc = self.run(['--symmetric'],
                        create_fhs=['stdin', 'stdout'])
        proc.handles['stdin'].write(string)
        proc.handles['stdin'].close()
        output = proc.handles['stdout'].read()
        proc.handles['stdout'].close()
        proc.wait()
        return output

    def armor_string(self, string):
        self.options.recipients = []
        proc = self.run(['--enarmor'],
                        create_fhs=['stdin', 'stdout'])
        proc.handles['stdin'].write(string)
        proc.handles['stdin'].close()
        output = proc.handles['stdout'].read()
        proc.handles['stdout'].close()
        proc.wait()
        return output

    def clearsign_string(self, string, signer, passphrase):
        self.options.recipients = []
        self.options.default_key = signer
        self.passphrase = passphrase
        proc = self.run(['--clearsign'],
                        create_fhs=['stdin', 'stdout'])
        proc.handles['stdin'].write(string)
        proc.handles['stdin'].close()
        output = proc.handles['stdout'].read()
        proc.handles['stdout'].close()
        proc.wait()
        return output

def emit_alist(f, d):
    f.write("(\n")
    keys = d.keys()
    keys.sort()
    for k in keys:
        if d[k] == None:
            f.write("(%s . nil)\n" % (k))
        else:
            f.write("(%s . \"%s\")\n" % (k, d[k]))
    f.write(")\n")

encryptor = Encrypt()
def encrypt(to, plaintext):
    return encryptor.encrypt_string(plaintext, [to])
def signencrypt(to, signer, plaintext):
    return encryptor.signencrypt_string(plaintext, signer, signer, [to])
def sign(signer, plaintext):
    return encryptor.sign_string(plaintext, signer, signer)
def sym(passphrase, plaintext):
    return encryptor.sym_string(plaintext, passphrase)
def armor(plaintext):
    # note: this isn't very useful, you must use 'gpg --dearmor' to
    # extract the result, not --decrypt. It also has a different banner.
    return encryptor.armor_string(plaintext)
def clearsign(signer, plaintext):
    return encryptor.clearsign_string(plaintext, signer, signer)

#def encrypt(to, plaintext):
    #cmd = "gpg --armor --homedir %s --recipient %s --batch --encrypt" % (homedir, to)
    #(stdout,stdin) = popen2.popen2(cmd)
    #stdin.write(plaintext)
    #stdin.close()
    #crypttext = stdout.read()
    #return crypttext

class TestCase:
    def __init__(self, filename):
        self.filename = filename
        self.plaintext = self.make_plaintext()
        self.d = {}
        self.d['name'] = self.filename
        self.d['plaintext'] = self.plaintext
        self.d['error'] = None
        self.d['signature_status'] = None

    def make_plaintext(self):
        return "This is a plaintext message\n"
    def date_string(self):
        # mc-gpg.el takes the YYYY-MM-DD date string from the SIG_ID
        # status-fs line and delivers it to the user. This appears to be the
        # GMT date of the signature. Extract the same thing here so we can
        # tell the test harness what to expect. This needs to run on the
        # same day as the gpg invocation used to create the signature, but
        # this whole program only takes a few seconds to execute.
        return time.strftime("%Y-%m-%d", time.gmtime()) # GMT

    def encrypted_fields(self, recip):
        # run this after signed_fields so [signature_status] gets cleared
        self.d['encryption_id'] = "0x%s" % id[recip][-16:]
        self.d['passphrase'] = recip
        if recip == "other" or recip == "unknown":
            self.d['error'] = "This message is not addressed to you"
            self.d['plaintext'] = None
            self.d['signature_status'] = None

    def symencrypted_fields(self):
        self.d['encryption_id'] = "***** CONVENTIONAL *****"
        self.d['passphrase'] = self.passphrase

    def signed_fields(self, signer):
        self.d['signature_status'] = ("Good signature from '%s <%s@test>' " + \
                                 "TRUST_%s made %s") % \
                                 (signer, signer,
                                  trustmap[signer], self.date_string())
        if signer == "unknown":
            self.d['signature_status'] = "cannot check signature " + \
                                         "from keyid %s" % \
                                         id[signer][-16:]
            # comes from 'cannot check signature' warning
    def process(self, testcasedir):
        self.make_crypttext()
        f = open(os.path.join(testcasedir, self.filename), "w")
        emit_alist(f, self.d)
        f.close()


class E_Case(TestCase):
    def __init__(self, filename, recip):
        TestCase.__init__(self, filename)
        self.recip = recip
        self.encrypted_fields(recip)
    def make_crypttext(self):
        self.d['crypttext'] = "\n" + encrypt(self.recip, self.plaintext)

class Latin1_E_Case(E_Case):
    # encrypt a message that contains ISO-Latin-1 characters
    def make_plaintext(self):
        return """
Usted tiene que mostrar esta evidencia escrita a un oficial de policía cuando
éste lo solicite así como a otro conductor o dueño do propiedad después de un
accidente vial."""

class EE_Case(E_Case):
    # encrypt to multiple recipients, some unknown
    def make_crypttext(self):
        self.d['crypttext'] = "\n" + \
                              encryptor.encrypt_string(self.plaintext,
                                                       ["unknown",
                                                        "other",
                                                        self.recip])
class ES_Case(TestCase):
    def __init__(self, filename, recip, signer):
        TestCase.__init__(self, filename)
        self.recip = recip
        self.signer = signer
        self.signed_fields(signer)
        self.encrypted_fields(recip)
    def make_crypttext(self):
        self.d['crypttext'] = "\n" + signencrypt(self.recip,
                                                 self.signer,
                                                 self.plaintext)

class S_Case(TestCase):
    def __init__(self, filename, signer):
        TestCase.__init__(self, filename)
        self.signer = signer
        self.signed_fields(signer)
    def make_crypttext(self):
        self.d['crypttext'] = "\n" + sign(self.signer, self.plaintext)

class CS_Case(TestCase):
    def __init__(self, filename, signer):
        TestCase.__init__(self, filename)
        self.signer = signer
        self.signed_fields(signer)
    def make_crypttext(self):
        self.d['crypttext'] = "\n" + clearsign(self.signer, self.plaintext)

class Latin1_CS_Case(CS_Case):
    def make_plaintext(self):
        return """
Usted tiene que mostrar esta evidencia escrita a un oficial de policía cuando
éste lo solicite así como a otro conductor o dueño do propiedad después de un
accidente vial.
"""

class SYM_Case(TestCase):
    def __init__(self, filename, passphrase):
        TestCase.__init__(self, filename)
        self.passphrase = passphrase
        self.symencrypted_fields()
    def make_crypttext(self):
        self.d['crypttext'] = "\n" + sym(self.passphrase, self.plaintext)

def make_cases():
    d = testcasedir
    E_Case("E.e1r", "owner1").process(d)
    E_Case("E.e2r", "owner2").process(d)
    E_Case("E.e3", "other").process(d)
    E_Case("E.e4", "unknown").process(d)
    ES_Case("ES.e1r.s1v", "owner1", "owner1").process(d)
    ES_Case("ES.e1r.s2v", "owner1", "trusted").process(d)
    ES_Case("ES.e1r.s3v", "owner1", "untrusted").process(d)
    ES_Case("ES.e1r.s4", "owner1", "unknown").process(d)
    ES_Case("ES.e3.s1v", "other", "owner1").process(d)
    ES_Case("ES.e4.s1v", "unknown", "owner1").process(d)
    S_Case("S.s1v", "owner1").process(d)
    S_Case("S.s2v", "trusted").process(d)
    S_Case("S.s3v", "untrusted").process(d)
    S_Case("S.s4", "unknown").process(d)
    SYM_Case("SE", "password").process(d)
    CS_Case("CS.s1v", "owner1").process(d)
    CS_Case("CS.s2v", "trusted").process(d)
    CS_Case("CS.s3v", "untrusted").process(d)
    CS_Case("CS.s4", "unknown").process(d)
    Latin1_CS_Case("CS.latin1.s1v", "owner1").process(d)
    Latin1_E_Case("E.latin1.e1r", "owner1").process(d)
    EE_Case("E.e1re3re4r", "owner1").process(d)

# get keyids
id = {}
for u in users:
    id[u] = get_keyid(u)

trustmap = {
    'owner1': "ULTIMATE",
    'owner2': "ULTIMATE",
    'other': "FULL",
    'unknown': "NONE",
    'trusted': "MARGINAL",
    'untrusted': "UNDEFINED",
    }

if __name__ == "__main__":
    if not os.path.isdir(testcasedir):
        os.mkdir(testcasedir)
    make_cases()


# todo:
#  hardcoded date in es()
#  cs(): corrupt the signatures by editing the crypttext after signing



#! /usr/bin/python

# This is a class which watches a maildir for new messages. It uses the
# linux dirwatcher API (if available) to look for new files. The
# .messageReceived method is invoked with the filename of the new message,
# relative to the top of the maildir (so it will look like "new/blahblah").

# this is an abstract base class. It must be subclassed by something to
# provide a delay function (which polls in the case that DNotify isn't
# available) and a way to safely schedule code to run after a signal handler
# has fired. See maildirgtk.py and maildirtwisted.py for forms that use the
# event loops provided by Gtk+ and Twisted.

try:
    from dnotify import DNotify
    have_dnotify = 1
except:
    have_dnotify = 0
import os, os.path

class Maildir:
    def __init__(self, basedir):
        self.basedir = basedir
        self.newdir = os.path.join(basedir, "new")
        self.files = []
        self.pollinterval = 10  # only used if we don't have DNotify
        self.running = 0

    def start(self):
        """You must run start to receive any messages."""
        if self.running:
            return
        self.running = 1
        if not os.path.isdir(self.basedir) or not os.path.isdir(self.newdir):
            raise "invalid maildir '%s'" % basedir
        # we must hold an fd open on the directory, so we can get notified
        # when it changes.
        if have_dnotify:
            self.dnotify = DNotify(self.newdir, self.dnotify_callback,
                                   [DNotify.DN_CREATE])
        self.poll()

    def startTimeout(self):
        raise NotImplemented
    def stopTimeout(self):
        raise NotImplemented
    def dnotify_callback(self):
        print "callback"
        self.poll()
        raise NotImplemented

    def stop(self):
        if have_dnotify:
            self.dnotify.remove()
            del(self.dnotify)
        else:
            self.stopTimeout()
        self.running = 0

    def poll(self):
        # see what's new
        for f in self.files:
            if not os.path.isfile(os.path.join(self.newdir, f)):
                self.files.remove(f)
        newfiles = []
        for f in os.listdir(self.newdir):
            if not f in self.files:
                newfiles.append(f)
        self.files.extend(newfiles)
        for n in newfiles:
            self.messageReceived(os.path.join("new",n))
        if not have_dnotify:
            self.startTimeout()

    def messageReceived(self, filename):
        """Called when a new file is noticed. Override it in subclasses."""
        print filename


def test1():
    m = Maildir("ddir")
    m.start()
    import signal
    while 1:
        signal.pause()

if __name__ == '__main__':
    test1()




#! /usr/bin/python

import popen2, commands, re, os, sys

# The stripped ring contains the following subset, with the given trust values:

# owner1 (public+secret) ultimate  u
# owner2 (public+secret) ultimate  u
# other (public only) marginal     m
# trusted (public only) do not trust n
# untrusted (public only) "don't know" q

# as a result, messages signed by these keys should have the following computed
# trust values:
#  owner1 TRUST_ULTIMATE u
#  owner2 TRUST_ULTIMATE u
#  other  TRUST_FULLY f
#  unknown none -
#  trusted TRUST_MARGINAL m
#  untrusted TRUST_NONE -

homedir = sys.argv[1]

def get_keyid(name):
    # now what keyid did it get?
    cmd = "gpg --homedir %s --with-colons --fingerprint %s" % (homedir,name)
    (s,out) = commands.getstatusoutput(cmd)
    if s != 0:
        return None
    r = re.search(r'^fpr:::::::::(\w+):',out, re.M)
    if r == None:
        print "problem, out '%s'" % out
        return None
    id = r.group(1) # long form
    return id

def make_trustdb(trust):
    print "updating trustdb"
    cmd = "gpg --homedir %s --import-ownertrust" % homedir
    (stdout,stdin) = popen2.popen2(cmd)
    for id in trust.keys():
        stdin.write("%s:%d:\n" % (id, trust[id]))
    stdin.close()


users = ("owner1", "owner2", "other", "trusted", "untrusted", "unknown")

# get keyids
id = {}
for u in users:
    id[u] = get_keyid(u)

# and assign trust values
trust = {
    id["owner1"]: 6,    # u
    id["owner2"]: 6,    # u
    id["other"]: 4,     # m
    id["trusted"]: 3,    # n
    id["untrusted"]: 2, # q
    }
if id["unknown"]:
    trust[id["unknown"]] = 6 # u

make_trustdb(trust)
os.system("gpg --homedir %s --check-sigs >/dev/null" % homedir)
os.system("gpg --homedir %s --check-trustdb" % homedir)



#! /usr/bin/python

# This is a class which watches a maildir for new messages. It uses the
# linux dirwatcher API (if available) to look for new files. The
# .messageReceived method is invoked with the filename of the new message,
# relative to the top of the maildir (so it will look like "new/blahblah").

# This version is implemented as a Twisted Python "Service". It uses the
# twisted Reactor to handle polling and signal safety.

from twisted.internet.app import ApplicationService
from twisted.internet import reactor
from maildir import Maildir

class MaildirTwisted(ApplicationService, Maildir):
    def __init__(self, serviceName, serviceParent, basedir):
        ApplicationService.__init__(self, serviceName)
        Maildir.__init__(self, basedir)
    def startService(self):
        self.serviceRunning = 1
        self.start()
        return None
    def stopService(self):
        self.stop()
        return None
    def startTimeout(self):
        self.timeout = reactor.callLater(self.pollinterval, self.poll)
    def stopTimeout(self):
        if self.timeout:
            self.timeout.cancel()
            self.timeout = None
    def dnotify_callback(self):
        # make it safe
        reactor.callFromThread(self.poll)


def test1():
    class MaildirTest(MaildirTwisted):
        def messageReceived(self, filename):
            print "changed:", filename
    from twisted.internet.app import Application
    #from maildir import MaildirTest
    # note that MaildirTest is really __main__.MaildirTest, which makes
    # maildir-test-shutdown.tap unusable
    app = Application("maildir-test")
    # add the service to the app
    m = MaildirTest("maildir", app, basedir="ddir")
    print "watching ddir/new/"
    app.run()
    print "done"

if __name__ == '__main__':
    test1()





#! /usr/bin/python

# This is a class which watches a maildir for new messages. It uses the
# linux dirwatcher API (if available) to look for new files. The
# .messageReceived method is invoked with the filename of the new message,
# relative to the top of the maildir (so it will look like "new/blahblah").

# This form uses the Gtk event loop to handle polling and signal safety

if __name__ == '__main__':
    import pygtk
    pygtk.require("2.0")

import gtk
from maildir import Maildir

class MaildirGtk(Maildir):
    def __init__(self, basedir):
        Maildir.__init__(self, basedir)
        self.idler = None
    def startTimeout(self):
        self.timeout = gtk.timeout_add(self.pollinterval*1000, self.doTimeout)
    def doTimeout(self):
        self.poll()
        return gtk.TRUE # keep going
    def stopTimeout(self):
        if self.timeout:
            gtk.timeout_remove(self.timeout)
            self.timeout = None
    def dnotify_callback(self):
        # make it safe
        self.idler = gtk.idle_add(self.idlePoll)
    def idlePoll(self):
        gtk.idle_remove(self.idler)
        self.idler = None
        self.poll()
        return gtk.FALSE

def test1():
    class MaildirTest(MaildirGtk):
        def messageReceived(self, filename):
            print "changed:", filename
    m = MaildirTest("ddir")
    print "watching ddir/new/"
    m.start()
    #gtk.main()
    # to allow the python-side signal handler to run, we must surface from
    # gtk (which blocks on the C-side) every once in a while.
    while 1:
        gtk.mainiteration() # this will block until there is something to do
    m.stop()
    print "done"

if __name__ == '__main__':
    test1()



#! /usr/bin/python

import fcntl, signal

class DNotify_Handler:
    def __init__(self):
        self.watchers = {}
        self.installed = 0
    def install(self):
        if self.installed:
            return
        signal.signal(signal.SIGIO, self.fire)
        self.installed = 1
    def uninstall(self):
        if not self.installed:
            return
        signal.signal(signal.SIGIO, signal.SIG_DFL)
        self.installed = 0
    def add(self, watcher):
        self.watchers[watcher.fd.fileno()] = watcher
        self.install()
    def remove(self, watcher):
        if self.watchers.has_key(watcher.fd.fileno()):
            del(self.watchers[watcher.fd.fileno()])
            if not self.watchers:
                self.uninstall()
    def fire(self, signum, frame):
        # this is the signal handler
        # without siginfo_t, we must fire them all
        for watcher in self.watchers.values():
            watcher.callback()

class DNotify:
    DN_ACCESS = fcntl.DN_ACCESS  # a file in the directory was read
    DN_MODIFY = fcntl.DN_MODIFY  # a file was modified (write,truncate)
    DN_CREATE = fcntl.DN_CREATE  # a file was created
    DN_DELETE = fcntl.DN_DELETE  # a file was unlinked
    DN_RENAME = fcntl.DN_RENAME  # a file was renamed
    DN_ATTRIB = fcntl.DN_ATTRIB  # a file had attributes changed (chmod,chown)

    handler = [None]

    def __init__(self, dirname, callback=None,
                 flags=[DN_MODIFY,DN_CREATE,DN_DELETE,DN_RENAME]):

        """This object watches a directory for changes. The .callback
        attribute should be set to a function to be run every time something
        happens to it. Be aware that it will be called more times than you
        expect."""

        if callback:
            self.callback = callback
        else:
            self.callback = self.fire
        self.dirname = dirname
        self.flags = reduce(lambda x, y: x | y, flags) | fcntl.DN_MULTISHOT
        self.fd = open(dirname, "r")
        # ideally we would move the notification to something like SIGRTMIN,
        # (to free up SIGIO) and use sigaction to have the signal handler
        # receive a structure with the fd number. But python doesn't offer
        # either.
        if not self.handler[0]:
            self.handler[0] = DNotify_Handler()
        self.handler[0].add(self)
        fcntl.fcntl(self.fd, fcntl.F_NOTIFY, self.flags)
    def remove(self):
        self.handler[0].remove(self)
        self.fd.close()
    def fire(self):
        print self.dirname, "changed!"

def test_dnotify1():
    d = DNotify(".")
    import time
    while 1:
        signal.pause()

def test_dnotify2():
    # create ./foo/, create/delete files in ./ and ./foo/ while this is
    # running. Notice how both notifiers are fired when anything changes;
    # this is an unfortunate side-effect of the lack of extended sigaction
    # support in Python.
    count = [0]
    d1 = DNotify(".")
    def fire1(count=count, d1=d1):
        print "./ changed!", count[0]
        count[0] += 1
        if count[0] > 5:
            d1.remove()
            del(d1)
    # change the callback, since we can't define it until after we have the
    # dnotify object. Hmm, unless we give the dnotify to the callback.
    d1.callback = fire1
    def fire2(): print "foo/ changed!"
    d2 = DNotify("foo", fire2)
    import time
    while 1:
        signal.pause()


if __name__ == '__main__':
    test_dnotify2()




#! /usr/bin/python

if __name__ == '__main__':
    import pygtk
    pygtk.require("2.0")

import gtk

import os, os.path, stat, time, string
from maildirgtk import MaildirGtk
from nntplib import NNTP, NNTP_PORT, NNTPError

# This file implements some classes which watch maildir-style directories
# and NNTP newsgroups. A message is placed in the 'source' maildir when the
# user runs an elisp function just before encrypting the message for a
# remailer chain. Messages that emerge from the chain as email will wind up
# in the 'dest' maildir, which should be on the receiving end of a .qmail or
# .forward file. Messages that emerge as news postings should appear in a
# newsgroup monitored by the NewsWatcherService.

# The Watcher class watches both the source directory and the target
# dir/groups. It can provide a list of all files in 'source' that are not
# matched by files in 'dest'.

class Message:
    def __init__(self, msgid, name, time, data):
        self.msgid = msgid # the number inserted by the elisp counter
        self.name = name
        self.time = time # when the message was sent / arrived
        self.data = data # contents of the mesage (inc headers)
    def __repr__(self):
        return "<Message #%s>" % self.msgid

class MessageWatcher:
    def __init__(self):
        self.msgs = {}
        self.checker = None

    def setChecker(self, checker):
        self.checker = checker
    def __getstate__(self):
        d = self.__dict__.copy()
        d['checker'] = None
        return d

    def parseMessage(self, filename, name, time, lines):
        # find messageid, build bodytext
        data = ''
        msgid = None
        key = "MailcryptRemailerMessageId="
        for line in lines:
            line = line.rstrip("\n")
            data = data + line + "\n"
            where = line.find(key)
            if where != -1:
                where = where + len(key)
                msgid = line[where:]
        if msgid == None:
            return
        msgid = int(msgid)
        m = Message(msgid, name, time, data)
        if self.msgs.get(msgid, None) != None:
            print "Hey, file %s duplicates msgid %s from file %s" % \
                  (filename, msgid, self.msgs[msgid].filename)
        self.msgs[msgid] = m
        if self.checker:
            self.checker(m)

    def deleteMessage(self, msgid):
        try:
            del self.msgs[msgid]
        except KeyError:
            pass

    def dump(self):
        print len(self.msgs)," msgids:"
        for m in self.msgs.values():
            print m.msgid, "%d bytes, file %s" % (len(m.data), m.name)
        print


class DirWatcher(MaildirGtk, MessageWatcher):
    """This object watches a single maildir for incoming messages. It scans
    them for the MailcryptRemailerMessageId number, and stashes them in the
    .msgids dictionary by id. It also gives them to a .checker function, if
    set. The Maildir class tracks files that have been seen already, and
    keeps a persistent list of them, so the checker will only be called once
    per message."""
    def __init__(self, basedir):
        MessageWatcher.__init__(self)
        MaildirGtk.__init__(self, basedir)

    # we inherit start() and stop() from Maildir via MaildirGtk

    def messageReceived(self, name):
        # name is relative to the directory being watched
        #print "messageReceived", self.basedir, name
        filename = os.path.join(self.basedir, name)
        mtime = os.stat(filename)[stat.ST_MTIME]
        fd = open(filename)
        self.parseMessage(filename, name, mtime, fd.readlines())
        fd.close()

    def dump(self):
        print "dir:", self.basedir
        MessageWatcher.dump(self)

class NewsWatcher(MessageWatcher):
    def __init__(self, server, groups,
                 user=None, pw=None, port=None, tag=None):
        MessageWatcher.__init__(self)
        self.server = server
        self.groups = groups
        self.nntp = None  # the NNTP connection object
        self.user = user
        self.pw = pw
        self.port = port
        self.tag = tag
        self.last = {}
        self.timeout = None
        self.pollInterval = 60
        self.debug = 0

    def __repr__(self):
        return "<NewsWatcher %s:%s (%s)>" % (self.server, self.port,
                                             ",".join(self.groups))
    def __getstate__(self):
        d = MessageWatcher.__getstate__(self)
        d['nntp'] = None # just in case
        return d

    def start(self):
        port = self.port
        if not port:
            port = NNTP_PORT
        self.nntp = NNTP(self.server, port, self.user, self.pw,
                         readermode=1)
        # only look for messages that appear after we start. Usenet is big.
        if not self.last: # only do this the first time
            for g in self.groups:
                resp, count, first, last, name = self.nntp.group(g)
                self.last[g] = int(last)
                if self.debug: print "last[%s]: %d" % (g, self.last[g])
        self.timeout = gtk.timeout_add(self.pollInterval*1000,
                                       self.doTimeout)

    def stop(self):
        self.nntp.quit()
        self.nntp = None
        if self.timeout:
            gtk.timeout_remove(self.timeout)
            self.timeout = None

    def doTimeout(self):
        self.poll()
        return gtk.TRUE # keep going

    def poll(self):
        #print "polling", self
        for g in self.groups:
            resp, count, first, last, name = self.nntp.group(g)
            for num in range(self.last[g]+1, int(last)+1):
                try:
                    resp, num, id, lines = self.nntp.article("%d" % num)
                except NNTPError:
                    continue
                name = "%s:%d" % (g, int(num))
                if self.debug: print "got", name
                if self.tag:
                    if not filter(lambda line, tag=tag: line.find(tag) != -1,
                                  lines):
                        continue
                self.parseMessage(name, name, time.time(), lines)
            self.last[g] = int(last)

class Watcher:
    def __init__(self):
        self.source = None
        self.dests = []
        self.watchers = []
    def checker(self, m):
        # something changed, somewhere. The message 'm' was added, but we don't
        # know where. Override this method.
        pass

    def addSource(self, source):
        self.source = source
        self.watchers.append(source)
    def addDest(self, dest):
        self.dests.append(dest)
        self.watchers.append(dest)
    def start(self):
        if self.source:
            self.source.setChecker(self.checker)
        for d in self.dests:
            d.setChecker(self.checker)
        for w in self.watchers:
            w.start()
    def stop(self):
        for w in self.watchers:
            w.stop()

    def poll(self):
        # force a poll
        self.source.poll()
        for d in self.dests:
            d.poll()

    def msgs(self):
        # return dict of sent messages, and list of received messages. msgid
        # is the key for both, and the Message is the value.
        src = self.source.msgs
        dst = {}
        for d in self.dests:
            dst.update(d.msgs)
        return (src, dst)
    def outstanding(self):
        """Return a list of tuples (srcmsg, time), where time is the absolute
        time the message was sent."""
        src, dst = self.msgs()
        outstanding = []
        for msgid in src.keys():
            if not dst.has_key(msgid):
                m = src[msgid]
                outstanding.append((m, m.time))
        outstanding.sort(lambda x,y: cmp(x[0].msgid, y[0].msgid))
        return outstanding
    def age(self, msgid):
        tx_time = self.txtime(msgid)
        if not tx_time:
            return None
        age = time.time() - tx_time
        return age
    def txtime(self, msgid):
        if not self.source.msgs.has_key(msgid):
            return None
        tx_time = self.source.msgs[msgid].time
        return tx_time

    def received(self):
        """Return a list of tuples (dstmsg, latency), where latency is in
        seconds."""
        received = []
        src, dst = self.msgs()
        for msgid in src.keys():
            if dst.has_key(msgid):
                txtime = src[msgid].time
                rxtime = dst[msgid].time
                received.append((dst[msgid], rxtime - txtime))
        received.sort(lambda x,y: cmp(x[0].msgid, y[0].msgid))
        return received
    def abandon(self, msgid):
        # give up on the message: remove it from the source list
        self.source.deleteMessage(msgid)
        # just in case, remove it from the dest lists too
        for d in self.dests:
            d.deleteMessage(msgid)
    def flush(self, msgid):
        return self.abandon(msgid)


def do_test(wclass):
    # watch two maildirs
    w = wclass()
    sdir = DirWatcher("sdir")
    w.addSource(sdir)
    ddir = DirWatcher("ddir")
    w.addDest(ddir)
    print "running"
    w.start()
    while 1:
        gtk.mainiteration()
    w.stop()
    print "finished"

def test1():
    class Test1Watcher(Watcher):
        def checker(self, m):
            self.source.dump()
            self.dests[0].dump()
            print
    do_test(Test1Watcher)
def test2():
    class Test2Watcher(Watcher):
        def checker(self, m):
            print "outstanding:", self.outstanding()
            print "complete:", self.received()
            print
    do_test(Test2Watcher)

if __name__ == '__main__':
    test2()


# TODO: make sure messages that are present at startup get counted too. For
# Maildir, this means doing the scan in start instead of __init__. Must
# decide about NewsWatcher, should probably do a scan at __init__ time,
# record last[] info, then drop connection and reestablish at start.

